// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use fayalite::prelude::*;

#[hdl(outline_generated)]
pub struct S<T, Len: Size, T2> {
    pub a: T,
    b: UInt<3>,
    pub(crate) c: ArrayType<UInt<1>, Len>,
    pub d: T2,
}

#[hdl(outline_generated)]
pub struct S3<const LEN: usize, T> {
    pub a: T,
    b: UInt<3>,
    pub(crate) c: Array<UInt<1>, LEN>,
    pub d: S<T, ConstUsize<LEN>, ()>,
}

#[hdl(outline_generated)]
pub enum E<T> {
    A,
    B(UInt<3>),
    C(T),
}

#[hdl(outline_generated)]
pub struct S2<T = ()> {
    pub v: E<T>,
}

// check that #[hdl] properly handles hygiene
macro_rules! types_in_macros {
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $A:ident, $B:ident, $C:ident, $D:ident, $E:ident, $F:ident) => {
        #[hdl]
        struct $F {}
        #[hdl]
        struct $A<$B, $C: Size, const $D: usize, $E = $F> {
            $a: $B,
            $b: UIntType<$C>,
            $c: SInt<$D>,
            $d: HdlOption<$E>,
            $e: $E,
            $f: $F,
        }
        #[allow(non_camel_case_types)]
        #[hdl]
        enum $B<$C: Size, const $D: usize, $E = $F> {
            $a($A<(), $C, $D, $E>),
            $b(UIntType<$C>),
            $c(SInt<$D>),
            $d,
            $e($E),
            $f($F),
        }
    };
    // ensure every identifier has different hygiene
    () => {
        types_in_macros!(a);
    };
    ($a:ident) => {
        types_in_macros!($a, b);
    };
    ($a:ident, $b:ident) => {
        types_in_macros!($a, $b, c);
    };
    ($a:ident, $b:ident, $c:ident) => {
        types_in_macros!($a, $b, $c, d);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident) => {
        types_in_macros!($a, $b, $c, $d, e);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, f);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, $f, A);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $A:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, $f, $A, B);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $A:ident, $B:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, $f, $A, $B, C);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $A:ident, $B:ident, $C:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, $f, $A, $B, $C, D);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $A:ident, $B:ident, $C:ident, $D:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, $f, $A, $B, $C, $D, E);
    };
    ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $A:ident, $B:ident, $C:ident, $D:ident, $E:ident) => {
        types_in_macros!($a, $b, $c, $d, $e, $f, $A, $B, $C, $D, $E, F);
    };
}

types_in_macros!();
